# Getestet mit Python 3.11.6 und PySide 6.5.3

import time
import math
import PySide6.QtWidgets as widgets
import PySide6.QtCore as core
import PySide6.QtGui as gui

class Puck:
    def __init__(self, pos: core.QPointF, touchPos: core.QPointF, id: int):
        self.positions = [pos] # the past few positions
        self.times = [time.perf_counter()] # the times at which the past few events arrived
        self.idNumber = id
        self.velocityX = 0.0
        self.velocityY = 0.0
        self.touchPosition = touchPos # where the puck has been touched, relative to its center

class MeinWidget(widgets.QWidget):
    def __init__(self, parent : widgets.QWidget|None = None):
        super().__init__(parent)
        self.setCursor(core.Qt.CursorShape.BlankCursor)
        self.setAttribute(core.Qt.WidgetAttribute.WA_AcceptTouchEvents)
        self._colors = [core.Qt.GlobalColor.black, core.Qt.GlobalColor.red, core.Qt.GlobalColor.green, core.Qt.GlobalColor.blue, core.Qt.GlobalColor.yellow, core.Qt.GlobalColor.magenta, core.Qt.GlobalColor.gray]
        self._radius = 40.0

        self._pucks : list[Puck] = []

        self._timer = core.QTimer()
        self._timer.timeout.connect(self.animate)
        self._refreshRate = 50 # milliseconds
        self._timer.start(self._refreshRate)

    def paintEvent(self, event: gui.QPaintEvent) -> None:
        super().paintEvent(event)
        painter = gui.QPainter(self)
        n = 0
        for puck in self._pucks:
            c = self._colors[n % len(self._colors)]
            n += 1
            painter.setBrush(c)
            painter.setPen(core.Qt.PenStyle.NoPen)
            painter.drawEllipse(puck.positions[-1], self._radius, self._radius)
        painter.end()

    def event(self, event: core.QEvent) -> bool:
        if isinstance(event, gui.QTouchEvent):
            for touchPoint in event.points():
                print(touchPoint.state(), touchPoint.pos(), touchPoint.id())
                if touchPoint.state() == gui.QEventPoint.State.Pressed:
                    newPuckToBeCreated = True
                    for puck in self._pucks:
                        dx = touchPoint.pos().x() - puck.positions[0].x() # Here, puck.positions contains exactly one entry
                        dy = touchPoint.pos().y() - puck.positions[0].y()
                        distance = math.sqrt(dx**2 + dy**2)          
                        if distance < 2.0 * self._radius: # don't create new pucks too close to existing ones
                            newPuckToBeCreated = False
                        if distance < self._radius:
                            puck.touchPosition = core.QPointF(dx, dy)
                            puck.idNumber = touchPoint.id()
                            break
                    if newPuckToBeCreated:
                        p = Puck(touchPoint.pos(), core.QPointF(), touchPoint.id())
                        self._pucks.append(p)
                elif touchPoint.state() == gui.QEventPoint.State.Updated:
                    for puck in self._pucks:
                        if puck.idNumber == touchPoint.id():
                            puck.times.append(time.perf_counter())
                            puck.positions.append(core.QPointF(touchPoint.pos().x() - puck.touchPosition.x(), touchPoint.pos().y() - puck.touchPosition.y()))
                            if len(puck.times) > 10: # do not store more than the past ten
                                del puck.positions[0]
                                del puck.times[0]
                elif touchPoint.state() == gui.QEventPoint.State.Released:
                    for puck in self._pucks:
                        if puck.idNumber == touchPoint.id():
                            dt = puck.times[-1] - puck.times[0] + 0.001 # small offset to never divide by zero
                            puck.velocityX = (puck.positions[-1].x() - puck.positions[0].x()) / dt
                            puck.velocityY = (puck.positions[-1].y() - puck.positions[0].y()) / dt
                            puck.idNumber = -1
                            if len(puck.positions) > 0: # only leave the latest timed position
                                del puck.positions[0:-1]
                                del puck.times[0:-1]
            self.update()
            return True
        else:
            return super().event(event)
        
    def animate(self) -> None:
        for puck in self._pucks:
            if puck.idNumber == -1: # is not under finger
                x = puck.positions[-1].x() + puck.velocityX * self._refreshRate / 1000.0
                y = puck.positions[-1].y() + puck.velocityY * self._refreshRate / 1000.0
                # bounce off borders
                if x < self._radius:
                    x = self._radius
                    puck.velocityX = -puck.velocityX
                elif x > self.width() - self._radius:
                    x = self.width() - self._radius
                    puck.velocityX = -puck.velocityX
                if y < self._radius:
                    y = self._radius
                    puck.velocityY = -puck.velocityY
                elif y > self.height() - self._radius:
                    y = self.height() - self._radius
                    puck.velocityY = -puck.velocityY
                puck.positions[-1] = core.QPointF(x, y)
        self.update()
